(*
 *	dashc, client emulator for DASH video streaming
 *	Copyright (c) 2016-2018, Aleksandr Reviakin, University College Cork
 *                     2018, Andrew Nash, Benjamin Provan-Bessell, MISL Summer of Code 2018
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version 2
 *	of the License, or (at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 *	02110-1301, USA.
*)
open Sys
open Core
open Async
open Stream
open Cohttp_async
open Representation
open Segm_result
open Adapt_algo
let supported_algorithms = ["conv"; "bba-0"; "bba-1"; "bba-2"; "arbiter"]
let algorithms_with_required_chunk_sizes = ["bba-1"; "bba-2"; "arbiter"]

(*depending on the users choice different amounts of data will be displayed*)
let log_columns_ext = ["Seg_#"; "Arr_time"; "Del_Time"; "Stall_Dur";
"Rep_Level"; "Del_Rate"; "Act_Rate"; "Byte_Size"; "Buff_Level";  "RTT   "; "  Codec"; " Width"; " Height "; "FPS "; "Seg_Dur"; " Start"; " P.1203 QoE"]
let log_columns = ["Seg_#"; "Arr_time"; "Del_Time"; "Stall_Dur";
"Rep_Level"; "Del_Rate"; "Act_Rate"; "Byte_Size"; "Buff_Level";  "RTT   "]

(*track current segment number*)
let segment_count = ref 0
let current_bitrate = ref "0"

(*Log all data for this run in a temporary directory based on the current timestamp*)
let tmp_dir = Time.to_filename_string (Time.now ()) ~zone:Time.Zone.utc
let list_dir = String.split tmp_dir ~on:'.'
let dir = String.concat ~sep: "" list_dir
let seg_dir = "./" ^ dir
let tmp_seg_dir = "./" ^ dir

(*global variable to track metrics for summary statistics*)
let total_rtt = ref 0.
let min_rtt = ref 10000000000000.
let max_rtt = ref 0.
let num_segs_done = ref 0
let total_del_rate = ref 0
let last_qoe = ref 0.
let kb_downloaded = ref 0.
let total_rep_level = ref 0
let min_rep_level = ref 100000000000000000
let max_rep_level = ref 0
let min_del_rate = ref 1000000000000.
let max_del_rate = ref 0.
let min_res = ref 1000000000
let min_w = ref 0
let max_w = ref 0
let max_res = ref 0

(*global variables to track warnings - will only warn the user about an issue twice*)
let firstwarn = ref true
let firstwarn_res = ref true
let warned_time = ref false

(*keep track of stall data to adjust start times and track summary stall statistics*)
let stalled = ref 0.
let no_stalls = ref 0
let time_stalled = ref 0
(*start_stalled tracks the start time of the current segment, accounting for any stalls that took place*)
let start_stalled = ref 0.
let stop_qoe = ref false

(*global variables to track resolutions - could be improved by using map/hash data structure*)
let r180 = ref 0
let r432 = ref 0
let r216 = ref 0
let r240 = ref 0
let r288 = ref 0
let r360 = ref 0
let r414 = ref 0
let r720 = ref 0
let r144 = ref 0
let r576 = ref 0
let r1080 = ref 0
let r2160 = ref 0
let h180 = ref 0
let h216 = ref 0
let h240 = ref 0
let h144 = ref 0
let h576 = ref 0
let h432 = ref 0
let h288 = ref 0
let h360 = ref 0
let h414 = ref 0
let h720 = ref 0
let h1080 = ref 0
let h2160 = ref 0

(*
let first_line_of_file name =
  bind (Reader.open_file name) (fun file ->
    bind (Reader.read_line file) (fun line ->
      return line
    )
  )
*)

let read_line i = try Some (input_line i) with End_of_file -> None

let lines_from_files filename =
  let rec lines_from_files_aux i acc = match (read_line i) with
    | None -> List.rev acc
    | Some s -> lines_from_files_aux i (s :: acc) in
  lines_from_files_aux (open_in filename) []

(*does the buffer contain enough free memory to contain another whole segment*)
let loss_before = ref 0

let is_buffer_full buffer_size maxbuf segment_duration =
  buffer_size > (maxbuf -. float_of_int segment_duration)

let put_in_file name text =
  Writer.save name ~contents:text

let next_repr_number representations results last_segment_index = function
  | Conv ->
    Conv.next_representation_level ~representations:representations ~results:results
  | BBA_0 algo ->
    BBA_0.next_representation_level
    ~algo:algo ~representations:representations ~results:results
  | BBA_1 algo ->
    BBA_1.next_representation_level
    ~algo:algo ~representations:representations ~results:results last_segment_index
  | BBA_2 algo ->
    BBA_2.next_representation_level
    ~algo:algo ~representations:representations ~results:results last_segment_index
  | ARBITER algo -> ARBITER.next_representation_level
    ~algo:algo ~representations:representations ~results:results last_segment_index


let try_request_next_chunk ?conn link_to_next_chunk extra_cols =
  try_with (fun () ->
  Client.get ?conn:conn (Uri.of_string link_to_next_chunk)
    >>| fun (resp, body) ->
      segment_count:=(!segment_count)+1;

      (* extra *)
      let _ = if (extra_cols=true) then (
      (*write segment chunks to files as they are downloaded*)
      (*these files are renamed later on to allow for more effective post-processing*)
      (*the reason that they are renamed later rather than named here is that there is
       more data available to create a more detailed nameing system*)
      let file = seg_dir ^ "/seg-" ^ string_of_int(!segment_count) ^ ".m4s" in
      let oc = Out_channel.create ~append:true @@ file in
      (*assuming that the async.pipe acts as a FIFO queue-tpye data structure*)
      let pipe = Cohttp_async.Body.to_pipe body in
      (* Convert the received file into an Async Pipe through which we can iterate and store the partial segment chunks*)
      let rec read_from_pipe pipe =
        Async.Pipe.read pipe>>=
          function
          | `Eof ->
              Out_channel.close oc;
              return ()
          | `Ok x ->  Out_channel.output_string oc @@ Printf.sprintf "%s" x;
                      return ()
                      >>=
                          fun () -> read_from_pipe pipe
      in
      (* recursively read each segment piece from the Pipe *)
    ignore(read_from_pipe pipe);
    )
    else () in
      Some (resp, body)
    )
  >>| function
  | Ok resp_and_body -> resp_and_body
  | Error e -> begin
    match (String.is_substring (Exn.to_string e) ~substring:"Connection closed by remote host") with
    | true -> None
    | false -> failwith @@ Exn.to_string e
    end

let rec playback
    ?conn
    ?outc
    ~root_link
    ~representations
    ~results
    ~absolute_start_time
    ~buffer_size
    ~segment_number
    ~initb
    ~maxbuf
    ~adapt_method
    ~last_segment_index
    ~segment_duration
    ~link
    ~persist
    ~extra_cols
    ~no_segs
    ~max_rep_height
     =
  match is_buffer_full buffer_size maxbuf segment_duration with
  | true ->
    (*Pause for 0.1 seconds due to full buffer and reurse*)
    Clock.after (Time.Span.of_sec 0.1) >>= fun () ->
    playback
      ?conn:conn
      ?outc:outc
      ~root_link:root_link
      ~representations:representations
      ~results:results
      ~absolute_start_time:absolute_start_time
      ~buffer_size:(buffer_size -. 0.1)
      ~segment_number:(segment_number)
      ~initb:initb
      ~maxbuf:maxbuf
      ~adapt_method:adapt_method
      ~last_segment_index:last_segment_index
      ~segment_duration:segment_duration
      ~link:link
      ~persist:persist
      ~extra_cols:extra_cols
      ~no_segs:no_segs
      ~max_rep_height:max_rep_height
  | false ->
    let start = Time.now () in
    let next_repr_number = next_repr_number representations results last_segment_index adapt_method in
    let next_repr = Hashtbl.find_exn representations next_repr_number in

  (*determine codec used*)
    let is_h264 = if next_repr.codec.[0]='a' && next_repr.codec.[1]='v' && next_repr.codec.[2]='c' then true else false in
    let is_h265 = if next_repr.codec.[0]='h' && next_repr.codec.[1]='e' && next_repr.codec.[2]='v' then true else false in
    let is_vp9  = if next_repr.codec.[0]='v' && next_repr.codec.[1]='p'                            then true else false in
    let is_av1  = if next_repr.codec.[0]='a' && next_repr.codec.[1]='v' && next_repr.codec.[2]='1' then true else false in

    (*
    this needs to be at the start of play, before we download the next segment
    if we have reached the final segment specifiec by the user in -no-segs, print summary and exit
    *)
    let _ = if (!num_segs_done = no_segs)
    then (
      (*determine most common resolution using gloabal variables*)
      let mode = ref 0 in
      let count = ref 0 in
      let width = ref 0 in
      let _ = if (!r180 > !count)  then (mode:=180;  count:=!r180;  width:=!h180)   else () in
      let _ = if (!r216 > !count)  then (mode:=216;  count:=!r216;  width:=!h216)   else () in
      let _ = if (!r288 > !count)  then (mode:=288;  count:=!r288;  width:=!h288)   else () in
      let _ = if (!r144 > !count)  then (mode:=144;  count:=!r144;  width:=!h144)   else () in
      let _ = if (!r576 > !count)  then (mode:=576;  count:=!r576;  width:=!h576)   else () in
      let _ = if (!r360 > !count)  then (mode:=360;  count:=!r360;  width:=!h360)   else () in
      let _ = if (!r432 > !count)  then (mode:=432;  count:=!r432;  width:=!h432)   else () in
      let _ = if (!r240 > !count)  then (mode:=240;  count:=!r240;  width:=!h240)   else () in
      let _ = if (!r414 > !count)  then (mode:=414;  count:=!r414;  width:=!h414)   else () in
      let _ = if (!r720 > !count)  then (mode:=720;  count:=!r720;  width:=!h720)   else () in
      let _ = if (!r1080 > !count) then (mode:=1080; count:=!r1080; width:=!h1080)  else () in
      let _ = if (!r2160 > !count) then (mode:=2160; count:=!r2160; width:=!h2160)  else () in

      if (extra_cols) then (
      (*output statistics to summary file and standard output*)
      (*some metrics which are stated as KB or KB/s may have been divided by 1000 rather than 1024*)
      (*this is done in order to better reflect the data displayed in the data in the outputted columns*)
      let summary_file = Out_channel.create ~append:true @@ seg_dir ^ "/summary.log" in

      (*re-print warnings*)
      let _ = if (!firstwarn=false) then (printf "\nWARNING: codec not compatible with P.1203\n";Out_channel.output_string summary_file @@ Printf.sprintf "\nWARNING: codec not compatible with P.1203\n";) else () in
      let _ = if (!firstwarn_res=false) then (printf "\nWARNING: resolution has surpassed the maximum supported by P.1203\n";Out_channel.output_string summary_file @@ Printf.sprintf "\nWARNING: resolution has surpassed the maximum supported by P.1203\n") else () in
      let _ = if (!warned_time=true) then (stop_qoe:=true; printf "Warning: the stream has exceeded the ITU P.1203 time limit (5 mins) QoE values will be dropped\n";Out_channel.output_string summary_file @@ Printf.sprintf "Warning: the stream has exceeded the ITU P.1203 time limit (5 mins) QoE values will be dropped\n") else () in

      let _ = printf "\nDownloaded %d segments (%d-second %s) totalling    %.2f MB\n" !num_segs_done segment_duration (if (is_h264) then ("h264") else (if (is_h265) then ("h265") else (if (is_vp9) then ("vp9") else ("av1")))) (!kb_downloaded /. 1000.) in
      Out_channel.output_string summary_file @@ Printf.sprintf "\nDownloaded %d segments (%d-second %s) totalling    %.2f MB\n" !num_segs_done segment_duration (if (is_h264) then ("h264") else (if (is_h265) then ("h265") else (if (is_vp9) then ("vp9") else ("av1")))) (!kb_downloaded /. 1000.);

      let _ = if ((!stop_qoe = true && !warned_time=false) || extra_cols=false) then () else (printf  "P.1203 Quality of Experience:\t\t\t   %.4f\n" !last_qoe;
      Out_channel.output_string summary_file @@ Printf.sprintf  "P.1203 Quality of Experience:\t\t\t   %.4f\n" !last_qoe) in

      let _ = (printf "Round Trip Time (RTT) min/avg/max:\t\t   %.3f/%.3f/%.3f\n" !min_rtt (!total_rtt /. float_of_int (!num_segs_done)) !max_rtt) in
      Out_channel.output_string summary_file @@ Printf.sprintf "Round Trip Time (RTT) min/avg/max:\t\t   %.3f/%.3f/%.3f\n" !min_rtt (!total_rtt /. float_of_int (!num_segs_done)) !max_rtt;

      let _ = (printf "Delivery rate min/avg/max:\t\t           %.3f/%.3f/%.3f KB/s\n" (!min_del_rate) (float_of_int(!total_del_rate) /. float_of_int(!num_segs_done))) (!max_del_rate) in
      Out_channel.output_string summary_file @@ Printf.sprintf "Delivery rate min/avg/max:\t\t           %.3f/%.3f/%.3f KB/s\n" (!min_del_rate) (float_of_int(!total_del_rate) /. float_of_int(!num_segs_done)) (!max_del_rate);

      let _ = (printf "Number of stalls/Total stall time:\t\t   %d/%.3f\n" !no_stalls (float_of_int (!time_stalled) /. 1000.) ) in
      Out_channel.output_string summary_file @@ Printf.sprintf "Number of stalls/Total stall time:\t\t   %d/%.3f\n" !no_stalls (float_of_int (!time_stalled) /. 1000.) ;

      let _ = printf "Representation level min/avg/max:\t\t   %.3f/%.3f/%.3f KB/s\n"  ((float_of_int !min_rep_level) /. 1000.) ((float_of_int (!total_rep_level) /. float_of_int (!num_segs_done))/.1000. ) ((float_of_int !max_rep_level) /. 1000.) in
      Out_channel.output_string summary_file @@ Printf.sprintf "Representation level min/avg/max:\t\t   %.3f/%.3f/%.3f KB/s\n" ((float_of_int !min_rep_level) /. 1000.) (float_of_int (!total_rep_level) /. float_of_int (!num_segs_done)/.1000. ) ((float_of_int !max_rep_level) /. 1000.);

      let _ = printf "Resolution min/mode/max:\t\t\t   %dx%d/%dx%d/%dx%d \n" (!min_w) (!min_res) (!width) (!mode) (!max_w) (!max_res) in
      Out_channel.output_string summary_file @@ Printf.sprintf "Resolution min/mode/max:\t\t\t   %dx%d/%dx%d/%dx%d \n" (!min_w) (!min_res) (!width) (!mode) (!max_w) (!max_res);

      Out_channel.close summary_file;
      ignore(exit 0);ignore()
    )
    )
    else ( ignore()) in

    let link_to_next_chunk = (root_link ^ (link_of_media next_repr.media segment_number next_repr.codec)) in
	try_request_next_chunk ?conn:conn link_to_next_chunk extra_cols
    >>= fun resp_and_body -> begin
    match resp_and_body with
    | None -> begin
        open_connection link persist >>= fun conn ->
        try_request_next_chunk ?conn:conn link_to_next_chunk extra_cols
        >>= fun resp_and_body_2 ->
        match resp_and_body_2 with
        | None -> failwith "The second disconnect in a raw"
        | Some (resp, body) ->
          (* print_endline "New connection was opened"; *)
          return (resp, body, conn)
      end
    | Some (resp, body) ->  return (resp, body, conn)
	end
  >>= fun (resp, body, conn) ->
			body |> Cohttp_async.Body.drain >>= fun () ->
					let content_length = match resp |> Cohttp_async.Response.headers |> Cohttp.Header.get_content_range with
					| Some content_length -> Int64.to_int_exn content_length
					| None -> 0
					in
					let current_time = Time.now () in
					let time_diff = Time.diff current_time start in
					let diff = Time.Span.to_sec time_diff  in
					let initb = match buffer_size >= (initb *. float_of_int segment_duration) with
					| true -> 0.
					| false -> initb
					in
					let buf = match buffer_size > (initb *. float_of_int segment_duration) with
					| true -> begin
					match buffer_size <= diff with
					| true -> 0.
					| false -> buffer_size -. diff
					end
					| false -> buffer_size
					in
					let stall = match buffer_size > (initb *. float_of_int segment_duration)
					&& buffer_size < diff with
					(* check this condition for startup phaze *)
					| true -> int_of_float @@ (diff -. buffer_size) *. 1000.
					| false -> 0
          in
          (*add any new stall to total stall time - convert it to seconds*)
          stalled := (float_of_int stall) *. 0.001;

          (*call shell script to calculate the video's bitrate and generate a mode 0 json file for itu p.1203*)
          (*get an associated qoe value and write it to a file (temp.tmp)*)
          let _ = if (extra_cols) then (

            (* lets create the file we need for storting our P.1203 value *)
            let _ = Sys.command @@ "touch " ^ "./" ^ seg_dir ^ "/temp" ^ string_of_int (segment_number) ^ ".tmp" in
            (* lets create the P.1203 value via shell *)
            ignore (
              let command_string =

            "mv  " ^ seg_dir ^ "/seg-" ^ string_of_int (segment_number) ^ ".m4s " ^          seg_dir ^  "/seg-" ^ string_of_int (segment_number) ^ "-" ^   string_of_int(next_repr.width)
             ^ "x" ^ string_of_int(next_repr.height) ^ "-" ^ string_of_int(next_repr.bandwidth) ^ ".m4s " ^
             "&&" ^

             " ./true_byte_size_calc.sh " ^ seg_dir
             ^  "/seg-" ^ string_of_int (segment_number) ^ "-" ^   string_of_int(next_repr.width) ^ "x" ^ string_of_int(next_repr.height) ^ "-"
            ^ string_of_int(next_repr.bandwidth) ^ ".m4s " ^ "\"h264\" \"" ^
             string_of_float (float_of_int (segment_duration))
            ^ "\" " ^ "\"" ^ string_of_int (next_repr.framerate) ^ "\" " ^ "\"" ^   string_of_int(next_repr.width) ^ "x" ^ string_of_int(next_repr.height) ^ "\" "
            ^ "\"" ^ string_of_float(!start_stalled +. !stalled) ^ "\" \"" ^ string_of_int (segment_number)  ^ "\" \"" ^ string_of_float(((float_of_int (stall)) *. 0.001)) ^

            "\" > " ^ seg_dir ^ "/temp" ^ string_of_int (segment_number) ^ ".tmp" in

              (* run the command *)
            Sys.command( command_string
              )
            )
            (*  let ic = Scanf.Scanning.from_file (seg_dir ^ "/temp" ^ string_of_int (segment_number) ^ ".tmp") in
            let qoe = Scanf.bscanf ic "%s" (fun a -> a) in
            let _ = if (qoe="" || qoe="0.000000") then () else (last_qoe:=float_of_string qoe) in *)
          ) else () in

          (*
          (* print out some shell information *)
          let _ = if (extra_cols) then (printf "mv  %s/seg-%d.m4s %s/seg-%d-%dx%d-%d.m4s &&\n" seg_dir segment_number seg_dir segment_number next_repr.width next_repr.height next_repr.bandwidth ) in

          (* print out some shell information *)
          let _ = if (extra_cols) then (printf "./true_byte_size_calc.sh %s/seg-%d-%dx%d-%d.m4s \"h264\" \"%d\" \"%d\"  \"%dx%d\" \"%f\" \"%d\" \"%f\" > %s/temp%d.tmp\n" seg_dir segment_number next_repr.width next_repr.height next_repr.bandwidth segment_duration next_repr.framerate next_repr.width next_repr.height (!start_stalled +. !stalled) segment_number (((float_of_int (stall)) *. 0.001)) seg_dir segment_number ) in
          *)

          let file = seg_dir ^ "/temp" ^ string_of_int (segment_number) ^ ".tmp" in

          (* delay so we can check on the status of the file *)
          let rec file_loop () = match file_exists (file) with
            (*| false -> Thread.delay(0.1); printf "waiting for file   \n"; file_loop () *)
            | false -> Thread.delay(0.1); file_loop ()
            | true -> ()
          in
          (* lets loop over the file until it is created *)
          let _ = file_loop () in

          (* loop so we can make sure a P.1203 value exists for us to use *)
          let rec qoe_loop () =
            let ic = if (extra_cols) then (
            (*get the QoE value out of the file*)
                Scanf.Scanning.from_file (file)
              )
              else (
                Scanf.Scanning.from_file ("./configure")
              ) in
            let qoe = if (extra_cols) then (
                Scanf.bscanf ic "%s" (fun a -> a)
            ) else "" in
            (*check if the qoe is valid*)
            if (extra_cols) then ( if (qoe="" || qoe="0.000000") then ( Thread.delay(0.1); (* printf "waiting for file input   \n";*) qoe_loop () ) else (last_qoe:=float_of_string qoe; Scanf.Scanning.close_in ic;) ); in

          (* lets loop over the qoe file until it has value *)
          let _ = qoe_loop () in

          (*the start time of the segment, accounting for stalls - if the segment did not stall !stalled will be 0, so it will be incremented by the segment duration*)
          start_stalled := (!start_stalled) +. float_of_int (segment_duration) +. !stalled;
          (*reset the duration of the current stall*)
          stalled := 0.;

          (*issue 5 min timeout warning for ITU P.1203 - only happens if the stream has taken 5 minutes,
          QoE values have not already been dropped and this warning has not been issued already*)
           let _ = if (!num_segs_done * segment_duration > 300 && !stop_qoe = false && !warned_time = false) then (warned_time:=true;printf "Warning: the stream has exceeded the ITU P.1203 time limit (5 mins) QoE values will be dropped\n"; stop_qoe:=true) else () in

          (*increment the amount of data downloaded by the size of the last segment*)
          kb_downloaded:=!kb_downloaded +. ((float_of_int content_length) /. 1024.);

          (*set minimum and maximum resolutions, if necessary*)
          let _ = if (!min_res > next_repr.height) then (min_res:=next_repr.height; min_w:=next_repr.width) else  () in
          let _ = if (!max_res < next_repr.height) then (max_res:=next_repr.height; max_w:=next_repr.width) else  () in
          let h = next_repr.height in

          (*increment the relevant global variable for the current resolution*)
          let _ = if (h=180) then(r180 := !r180 + 1;h180:=next_repr.width) else (if (h=216) then(r216 := !r216 + 1;h216:=next_repr.width)
          else (if (h=288) then(r288 := !r288 + 1;h288:=next_repr.width) else (if (h=360) then(r360 := !r360 + 1;h360:=next_repr.width)
          else (if (h=414) then(r414 := !r414 + 1;h414:=next_repr.width) else (if (h=720) then(r720 := !r720 + 1;h720:=next_repr.width) else (if (h=1080)
          then(r1080 := !r1080 + 1;h1080:=next_repr.width) else (if (h=2160) then(r2160 := !r2160 + 1;h2160:=next_repr.width) else (if (h=240)
          then(r240 := !r240 + 1;h240:=next_repr.width) else (if (h=576) then(r576 := !r576 + 1;h576:=next_repr.width) else (if (h=2160)
           then(r144 := !r144 + 1;h144:=next_repr.width) else (if (h=432) then(r432 := !r432 + 1;h432:=next_repr.width) else ()))))))))))) in

          (*warn for invalid codec for ITU P.1203 -> anything other than h264*)
          let _ = if (is_h264) then () else (if (!firstwarn=true) then (stop_qoe:=true;firstwarn:=false; printf "WARNING: codec not compatible with P.1203. QoE values will be dropped.\n") else()) in

          (*Get approximate rtt based on time to download a chunk of the next chunk file (only a few kbits max) - ping does not work with some http servers, so this method is used*)
          (*
	  let _ = printf "%s\n" link_to_next_chunk in
	  *)
          let a = Unix.gettimeofday() in
          Client.get (Uri.of_string link_to_next_chunk) >>= fun (_, _) ->
          let rtt = Unix.gettimeofday()  -. a in

          (*set global rtt tracking data*)
          total_rtt:=!total_rtt +. rtt;
          let _ = if (rtt <. !min_rtt) then (min_rtt:=rtt;) else () in
          let _ = if (rtt >. !max_rtt) then (max_rtt:=rtt;) else () in

          (*Warn about surpassing P.1203 resolution - don't bother if time limit has been surpassed*)
          let _ = if (next_repr.width>1920 && !warned_time=false) then (if (!firstwarn_res=true) then (stop_qoe:=true;firstwarn_res:=false;printf "WARNING: resolution has surpassed the maximum supported by P.1203. QoE values will be dropped\n";) else ()) else () in

          (*set global stall data*)
          let _ = if (stall=0) then () else (no_stalls:=!no_stalls+1;time_stalled:=!time_stalled + stall;) in

          (*set global delivery rate data*)
          total_del_rate:= !total_del_rate + int_of_float(((float_of_int content_length *. 8. *. 1000.) /. float_of_int (int_of_float @@ Time.Span.to_us time_diff;)) *. 1000. /. 1000.);
          let d =((float_of_int content_length *. 8. *. 1000.) /. float_of_int (int_of_float @@ Time.Span.to_us time_diff;)) *. 1000. /. 1000. in
          let _ = if (!min_del_rate > d) then (min_del_rate:=d) else () in
          let _ = if (!max_del_rate < d) then (max_del_rate:=d) else () in

          (*set global representation rate data*)
          total_rep_level:=!total_rep_level+next_repr.bandwidth;
          let _ = if (!min_rep_level>next_repr.bandwidth) then (min_rep_level:=next_repr.bandwidth) else () in
          let _ = if (!max_rep_level<next_repr.bandwidth) then (max_rep_level:=next_repr.bandwidth) else () in

          (*create results object and pass it  on to segm_result.ml*)
          let results =
					{
						segment_number = segment_number;
						arrival_time =
						int_of_float @@ Time.Span.to_us (Time.diff current_time absolute_start_time);
						time_for_delivery = int_of_float @@ Time.Span.to_us time_diff;
						stall_dur = stall;
						representation_rate = next_repr.bandwidth;
						actual_representation_rate = 8 * content_length / segment_duration;
						segment_size = content_length;
						buffer_level_in_momentum = (buf +. float_of_int segment_duration);
						repr_level = next_repr_number;
						width = next_repr.width;
						height = next_repr.height;
						framerate = next_repr.framerate;
						codec = if (is_h264) then "h264" else (if (is_h265) then "h265" else (if (is_vp9) then "vp9" else (if (is_av1) then "av1" else next_repr.codec)));
						duration = float_of_int (segment_duration);
            rtt = rtt;
            qoe = if (!stop_qoe = true) then ("0.000000") else (string_of_float !last_qoe);
            (* qoe = if (!stop_qoe = true) then ("0.000000") else (qoe);*)
          } :: results

          in
          (* Scanf.Scanning.close_in ic; *)
          num_segs_done := !num_segs_done + 1;
					print_result ( List.hd_exn results ) extra_cols seg_dir None ;
					let () = match outc with
					| Some outc -> print_result (List.hd_exn results) extra_cols seg_dir (Some outc)
					| None -> ()
					in

					match segment_number >= last_segment_index with
					| true -> Deferred.unit
					| false ->
					playback
						?conn:conn
						?outc:outc
						~root_link:root_link
						~representations:representations
						~results:results
						~absolute_start_time:absolute_start_time
						~buffer_size:(buf +. float_of_int segment_duration)
						~segment_number:(segment_number + 1)
						~initb:initb
						~maxbuf:maxbuf
						~adapt_method:adapt_method
						~last_segment_index:last_segment_index
						~segment_duration:segment_duration
						~link:link
						~persist:persist
            ~extra_cols:extra_cols
            ~no_segs:no_segs
            ~max_rep_height:max_rep_height

(* Run a command and return its results as a string. *)
(*let read_process command =
  let buffer_size = 2048 in
  let buffer = Buffer.create buffer_size in
  let string = Bytes.create buffer_size in
  let in_channel = Unix.open_process_in command in
  let chars_read = ref 1 in
  while !chars_read <> 0 do
    chars_read := input in_channel string 0 buffer_size;
    Buffer.add_substring buffer string 0 !chars_read
  done;
  ignore (Unix.close_process_in in_channel);
  Buffer.contents buffer
  *)

(* Run a command and return its results as a list of strings,
   one per line. *)
   (*
let read_process_lines command =
  let lines = ref [] in
  let in_channel = Unix.open_process_in command in
  begin
    try
      while true do
        lines := input_line in_channel :: !lines
      done;
    with End_of_file ->
      ignore (Unix.close_process_in in_channel)
  end;
  List.rev !lines
*)

let check_input_algorithm_existence input_alg =
  match List.exists supported_algorithms ~f:(fun x -> x = input_alg) with
  | true -> ()
  | false ->
    failwith "The chosen adaptation algorithm is not supported, please check help"

let create_log_channel logname log_folder v_number r_number = function
  | true ->
    begin match logname, v_number with
    | "now", _ ->
      begin match log_folder with
      | "" ->
        Some (Out_channel.create @@
          Time.to_filename_string (Time.now ()) ~zone:Time.Zone.utc)
      | _ ->
        Some (Out_channel.create
          (log_folder ^ "/" ^  Time.to_filename_string (Time.now ()) ~zone:Time.Zone.utc))
      end
    | _, "-V99" ->
      begin match log_folder with
      | "" -> Some (Out_channel.create logname)
      | _ -> Some (Out_channel.create (log_folder ^ "/" ^ logname))
      end
    | _, _ ->
      begin match log_folder with
      | "" ->
        Some (Out_channel.create
          ("trace-" ^ Time.to_filename_string (Time.now ())
          ~zone:Time.Zone.utc ^ "-" ^ logname ^ v_number ^ r_number ^ ".res"))
      | _ ->
        Some (Out_channel.create
          (log_folder ^ "/" ^ "trace-" ^ Time.to_filename_string (Time.now ())
          ~zone:Time.Zone.utc ^ "-" ^ logname ^ v_number ^ r_number ^ ".res"))
      end
    end
  | false -> None

let check_requirement_for_chunk_sizes input_alg =
  List.exists algorithms_with_required_chunk_sizes ~f:(fun x -> x = input_alg)

let chunk_sizes
    adapt_alg
    segm_size_from_file
    link representations
    last_segment_index
    root_link
    segmlist_mpd
    conn =
  match check_requirement_for_chunk_sizes adapt_alg with
  | true ->
    let chunk_sizes_per_repr = match segm_size_from_file with
    | "local" ->
        read_segment_size_file
          ?remote_string:None
          ~link:link
          ~number_of_representations:(Hashtbl.length representations)
          ~last_segment_index:last_segment_index
    | "remote" ->
        let segm_list_remote_link =
          root_link ^ "/" ^ "segmentlist_" ^ segmlist_mpd ^ ".txt" in
        Client.get ?conn:conn (Uri.of_string segm_list_remote_link)
        >>= fun (_, body) ->
        body |> Cohttp_async.Body.to_string >>= fun body_string ->
        read_segment_size_file
          ~remote_string:body_string
          ~link:link
          ~number_of_representations:(Hashtbl.length representations)
          ~last_segment_index:last_segment_index
    | _ ->
        download_chunk_sizes_per_repr
          ?conn:None
          ~root_link:root_link
          ~representations:representations
          ~last_segment_index:last_segment_index
    in
    chunk_sizes_per_repr >>| fun chunk_sizes_per_repr ->
    let average_chunk_size_per_repr =
      calculate_average_chunk_size_per_repr chunk_sizes_per_repr in
    (chunk_sizes_per_repr, average_chunk_size_per_repr)
  | false -> return (Hashtbl.Poly.create ~size:10 (), [])

let new_adapt_method chunk_sizes_per_repr average_chunk_size_per_repr = function
  | "conv" -> return Conv
  | "bba-0" -> return @@ BBA_0 { BBA_0.maxb = 240. }
  | "bba-1" ->
      return @@ BBA_1 {
        BBA_1.maxb = 240.;
        BBA_1.chunk_sizes_per_repr = chunk_sizes_per_repr;
        average_chunk_size_per_repr = average_chunk_size_per_repr
      }
  | "bba-2" ->
      return @@ BBA_2 {
        BBA_1.maxb = 240.;
        BBA_1.chunk_sizes_per_repr = chunk_sizes_per_repr;
        average_chunk_size_per_repr = average_chunk_size_per_repr
      }
  | "arbiter" ->
      return @@ ARBITER {
        ARBITER.maxb = 60.;
        ARBITER.chunk_sizes_per_repr = chunk_sizes_per_repr;
      }

  | _ -> failwith "The chosen adaptation algorithm is not supported"

let print_log_header_on_screen cols =
  List.iteri cols
  ~f:(fun idx x -> print_string x;
    match not @@ phys_equal idx (List.length cols - 1) with
    | true -> print_string "  "
    | false -> print_endline ""
  )

let print_log_header_into_file cols = function
  | Some outc -> List.iteri cols
    ~f:(fun idx x -> Out_channel.output_string outc x;
      match not @@ phys_equal idx (List.length cols - 1) with
      | true -> Out_channel.output_string outc "  "
      | false -> Out_channel.newline outc
    );
  | None -> ()


let get_max_muf maxbuf = function
  | BBA_0 algo -> int_of_float algo.BBA_0.maxb
  | BBA_1 algo -> int_of_float algo.BBA_1.maxb
  | BBA_2 algo -> int_of_float algo.BBA_1.maxb
  | ARBITER algo -> int_of_float algo.ARBITER.maxb
  | _ -> maxbuf

(* only create the log folder for P.1203 if needed *)
let create_extra_log_folder = function
  | true -> Sys.command @@ "mkdir -p ./" ^ dir ^ " && " ^ "touch " ^ "./" ^ dir ^ "/temp.tmp"
  | false -> Sys.command @@ ""

let run_client
    ~link
    ~adapt_alg
    ~initb
    ~maxbuf
    ~persist
    ~turnlogon
    ~logname
    ~v_number
    ~r_number
    ~log_folder
    ~last_segment_index
    ~segm_size_from_file
    ~extra_cols
    ~codec
    ~no_segs
    ~max_rep_height
     =
  try_with (fun () ->
      check_input_algorithm_existence adapt_alg;
      let _ = create_extra_log_folder extra_cols in
      begin
      match (log_folder = "") with
      | true -> return ()
      | false -> Deferred.ignore @@ Sys.command @@ "mkdir -p " ^ log_folder
      end

      >>= fun _ ->
      (* extra *)
      (*if (extra_cols) then *)
      (*let _ = Sys.command @@ "mkdir -p ./" ^ dir ^ " && " ^ "touch " ^ "./" ^ dir ^ "/temp.tmp" in*)

      let outc =
        create_log_channel
          logname (seg_dir ^ log_folder) ("-V" ^ v_number) ("-R" ^ r_number) turnlogon in
      open_connection link persist >>= fun conn ->
      Client.get ?conn:conn (Uri.of_string link) >>= fun (_, body) ->
      body |> Cohttp_async.Body.to_string >>= fun body ->
      let mpd = try Xml.parse_string body with | Dtd.Parse_error(_) -> printf "Error: No MPD file was found at the location you specified. Check if it has been moved or is unavailable and check that it consists of valid XML.\n"; ignore(exit 1);  Xml.parse_string "<a></a>" | _ -> Xml.parse_string body in
      (*print an error if there is a syntax error/unrecognised formating in the mpd file*)
      let representations : (int, representation) Hashtbl.t = try repr_table_from_mpd mpd codec max_rep_height extra_cols false
      with
      | Failure _ -> printf "Error: The MPD file you specified is invalid or incorrectly formatted.\n"; ignore(exit 1); repr_table_from_mpd mpd codec max_rep_height extra_cols false
      | _ -> repr_table_from_mpd mpd codec max_rep_height extra_cols false
      in
      (*if the hashtable is empty, then representation.ml did not return anythinig but did not throw an exception/error -
      this is caused by invalid codec resulting in all representations being skipped*)
      let _ = if (Hashtbl.is_empty representations) then (printf "Error: The codec you specified was not found in this mpd file\n";exit 1) else (return ())  in
      let segment_duration = (Hashtbl.find_exn representations 1).segment_duration in
      let last_segment_index =
        get_last_segment_index mpd segment_duration last_segment_index in
      let root_link, segmlist_mpd = String.rsplit2_exn link ~on:'/' in
      let root_link = root_link ^ "/" in
      chunk_sizes
        adapt_alg
        segm_size_from_file
        link
        representations
        last_segment_index
        root_link
        segmlist_mpd conn
      >>= fun (chunk_sizes_per_repr, average_chunk_size_per_repr) ->
      new_adapt_method chunk_sizes_per_repr average_chunk_size_per_repr adapt_alg
      >>= fun adapt_method ->
      if (extra_cols) then (print_log_header_on_screen log_columns_ext) else (print_log_header_on_screen log_columns);
      if (extra_cols) then(print_log_header_into_file log_columns_ext outc ) else (print_log_header_into_file log_columns outc );
      let maxbuf = get_max_muf maxbuf adapt_method in
      playback
        ?conn:conn
        ?outc:outc
        ~root_link:root_link
        ~representations:representations
        ~results:[]
        ~absolute_start_time:(Time.now ())
        ~buffer_size:0.
        ~segment_number:1
        ~initb:(float_of_int initb)
        ~maxbuf:(float_of_int maxbuf)
        ~adapt_method:adapt_method
        ~last_segment_index:last_segment_index
        ~segment_duration:segment_duration
        ~link:link
        ~persist:persist
        ~extra_cols:extra_cols
        ~no_segs:no_segs
        ~max_rep_height:max_rep_height
      >>= fun () ->
      (*get last segments data*)
      let next_repr = (Hashtbl.find_exn representations 1) in

      (*update resolution stats if necessary*)
      let mode = ref 0 in
      let count = ref 0 in
      let width = ref 0 in
      let _ = if (!r180 > !count) then (mode:=180; count:=!r180; width:=!h180)    else () in
      let _ = if (!r216 > !count) then (mode:=216; count:=!r216; width:=!h216)    else () in
      let _ = if (!r288 > !count) then (mode:=288; count:=!r288; width:=!h288)    else () in
      let _ = if (!r360 > !count) then (mode:=360; count:=!r360; width:=!h360)    else () in
      let _ = if (!r240 > !count) then (mode:=240; count:=!r240; width:=!h240)    else () in
      let _ = if (!r414 > !count) then (mode:=414; count:=!r414; width:=!h414)    else () in
      let _ = if (!r720 > !count) then (mode:=720; count:=!r720; width:=!h720)    else () in
      let _ = if (!r144 > !count) then (mode:=144; count:=!r144; width:=!h144)    else () in
      let _ = if (!r576 > !count) then (mode:=576; count:=!r576; width:=!h576)    else () in
      let _ = if (!r432 > !count) then (mode:=432; count:=!r432; width:=!h432)    else () in
      let _ = if (!r1080 > !count) then (mode:=1080; count:=!r1080; width:=!h1080) else () in
      let _ = if (!r2160 > !count) then (mode:=2160; count:=!r2160; width:=!h2160) else () in

      (*select codec*)
      let is_h264 = if next_repr.codec.[0]='a' && next_repr.codec.[1]='v' && next_repr.codec.[2]='c' then true else false in
      let is_h265 = if next_repr.codec.[0]='h' && next_repr.codec.[1]='e' && next_repr.codec.[2]='v' then true else false in
      let is_vp9  = if next_repr.codec.[0]='v' && next_repr.codec.[1]='p'                            then true else false in


      (*output statistics to summary file and standard output*)
      let _ = if (extra_cols) then (
      let summary_file = Out_channel.create ~append:true @@ seg_dir ^ "/summary.log" in

      let _ = if (!firstwarn=false) then (printf "\nWARNING: codec not compatible with P.1203\n";Out_channel.output_string summary_file @@ Printf.sprintf "\nWARNING: codec not compatible with P.1203\n";) else () in

      let _ = if (!firstwarn_res=false) then (printf "\nWARNING: resolution has surpassed the maximum supported by P.1203\n";Out_channel.output_string summary_file @@ Printf.sprintf "\nWARNING: resolution has surpassed the maximum supported by P.1203\n") else () in

      let _ = if (!warned_time=true) then (printf "Warning: the stream has exceeded the ITU P.1203 time limit (5 mins) QoE values will be dropped\n";Out_channel.output_string summary_file @@ Printf.sprintf "Warning: the stream has exceeded the ITU P.1203 time limit (5 mins) QoE values will be dropped\n") else () in

      let _ = printf "\nDownloaded %d segments (%d-second %s) totalling    %.2f MB\n" !num_segs_done segment_duration (if (is_h264) then ("h264") else (if (is_h265) then ("h265") else (if (is_vp9) then ("vp9") else ("av1")))) (!kb_downloaded /. 1000.) in
      Out_channel.output_string summary_file @@ Printf.sprintf "\nDownloaded %d segments (%d-second %s) totalling    %.2f MB\n" !num_segs_done segment_duration (if (is_h264) then ("h264") else (if (is_h265) then ("h265") else (if (is_vp9) then ("vp9") else ("av1")))) (!kb_downloaded /. 1000.);

      let _ = if ((!stop_qoe = true && !warned_time=false) || extra_cols=false) then () else (printf  "P.1203 Quality of Experience:\t\t\t   %.4f\n" !last_qoe;
      Out_channel.output_string summary_file @@ Printf.sprintf  "P.1203 Quality of Experience:\t\t\t   %.4f\n" !last_qoe) in

      let _ = (printf "Round Trip Time (RTT) min/avg/max:\t\t   %.3f/%.3f/%.3f\n" !min_rtt (!total_rtt /. float_of_int (!num_segs_done)) !max_rtt) in

      Out_channel.output_string summary_file @@ Printf.sprintf "Round Trip Time (RTT) min/avg/max:\t\t   %.3f/%.3f/%.3f\n" !min_rtt (!total_rtt /. float_of_int (!num_segs_done)) !max_rtt;

      let _ = (printf "Delivery rate min/avg/max:\t\t           %.3f/%.3f/%.3f KB/s\n" (!min_del_rate) (float_of_int(!total_del_rate) /. float_of_int(!num_segs_done))) (!max_del_rate) in

      Out_channel.output_string summary_file @@ Printf.sprintf "Delivery rate min/avg/max:\t\t           %.3f/%.3f/%.3f KB/s\n" (!min_del_rate) (float_of_int(!total_del_rate) /. float_of_int(!num_segs_done)) (!max_del_rate);

      let _ = (printf "Number of stalls/Total stall time:\t\t   %d/%.3f\n" !no_stalls (float_of_int (!time_stalled) /. 1000.) ) in
      Out_channel.output_string summary_file @@ Printf.sprintf "Number of stalls/Total stall time:\t\t   %d/%.3f\n" !no_stalls (float_of_int (!time_stalled) /. 1000.) ;

      let _ = printf "Representation level min/avg/max:\t\t   %.3f/%.3f/%.3f KB/s\n"  ((float_of_int !min_rep_level) /. 1000.) ((float_of_int (!total_rep_level) /. float_of_int (!num_segs_done))/.1000. ) ((float_of_int !max_rep_level) /. 1000.) in
      Out_channel.output_string summary_file @@ Printf.sprintf "Representation level min/avg/max:\t\t   %.3f/%.3f/%.3f KB/s\n" ((float_of_int !min_rep_level) /. 1000.) (float_of_int (!total_rep_level) /. float_of_int (!num_segs_done)/.1000. ) ((float_of_int !max_rep_level) /. 1000.);

      let _ = printf "Resolution min/mode/max:\t\t\t   %dx%d/%dx%d/%dx%d \n" (!min_w) (!min_res) (!width) (!mode) (!max_w) (!max_res) in
      Out_channel.output_string summary_file @@ Printf.sprintf "Resolution min/mode/max:\t\t\t   %dx%d/%dx%d/%dx%d \n" (!min_w) (!min_res) (!width) (!mode) (!max_w) (!max_res);

      Out_channel.close summary_file;
    ) else () in
      match outc with
      | Some outc ->  Out_channel.close outc; Deferred.unit
      | None -> Deferred.unit

  )
  >>| function
  | Ok () -> ()
  | Error e -> begin
    match (String.is_substring (Exn.to_string e) ~substring:"connection attempt timeout") with
      | true -> print_endline @@ "Connection attempt timeout (10 sec default) to " ^ link
      | false -> print_endline @@ Exn.to_string e
    end

let play =
  Command.async
    ~summary:"Dash simulation client"
    ~readme:(fun () -> "Usage example: ./dashc.native play http://10.0.0.1/bbb.mpd \
                        [-adapt conv] [-extended false] [-initb 2] [-maxbuf 60] [-persist true] \
                        [-turnlogon true] [-logname now] [-v 20] [-subfolder qwe] [-no_segs 999999999]")
    (
      let open Command.Let_syntax in
      [%map_open
        let link = anon ("link_to_mpd" %: string)
        and adapt_alg = flag "-adapt" (optional_with_default "conv" string)
          ~doc:" adaptation algorithm [conv; bba-0; bba-1; bba-2; arbiter]"
        and initb = flag "-initb" (optional_with_default 2 int)
          ~doc:" initial buffer in segments"
        and maxbuf = flag "-maxbuf" (optional_with_default 60 int)
          ~doc:" maximum buffer size"
        and persist = flag "-persist" (optional_with_default true bool)
          ~doc:" persistent connection"
        and turnlogon = flag "-turnlogon" (optional_with_default false bool)
          ~doc:" turn on logging to file"
        and logname = flag "-logname" (optional_with_default "now" string)
          ~doc:" name of the log file (\"now\" means usage of the current time stamp \
                and it is used by default)"
        and v_number = flag "-v" (optional_with_default "99" string)
          ~doc:" video number according to StreamTraceAnalysis.py"
        and r_number = flag "-r" (optional_with_default "0" string)
          ~doc:" the run (R) number for the trace file"
        and subfolder = flag "-subfolder" (optional_with_default "" string)
          ~doc:" subfolder for the file"
        and last_segment_index = flag "-lastsegmindex" (optional int)
          ~doc:" last segment index for the playback"
        and gensegmfile = flag "-gensegmfile" (optional_with_default false bool)
          ~doc:" generate segmentlist_%mpd_name%.txt file only \
                (it will be rewritten if exists)"
        and segm_size_from_file =
          flag "-segmentlist" (optional_with_default "head" string)
          ~doc:" get segment sizes from \n \
                local - local segmentlist_%mpd_name%.txt file \n \
                remote - download remote segmentlist_%mpd_name%.txt file \
                  from the same folder where mpd is located \n \
                head - get segment size by sending head requests \
                  before playing for each segment per representation \
                usage of head requests works only with non-persistent connection so far \
                  and it it set by default (for this operation)"
        and extra_cols =
           flag "-extended" (optional_with_default false bool)
           ~doc:" display extended columns of parameters \
                (codec, resolution, framerate, segment duration and Quality of Experience (QoE))"
        and no_segs =
          flag "-no-segs" (optional_with_default 9999999999 int)
         ~doc:" the number of segment files to download (999999999 (default) for all)"
        and pref_codec =
            flag "-preferred-codec" (optional_with_default "none" string)
              ~doc:" one of h264, h265, vp9 or av1. This will force dashc to use the selected codec when parsing a multi codec mpd file"
        and max_rep_height =
            flag "-max-rep-height" (optional_with_default 2160 int)
            ~doc:" the maximum resolution height to stream at"
        in

        (*check for bad codec argument*)
        let _ = if (pref_codec="h264" || pref_codec="h265" || pref_codec="vp9" || pref_codec="av1" || pref_codec="none") then () else (printf "Error: Unrecognised codec %s\n" pref_codec;ignore (exit 0);()) in
        begin
        fun () -> match gensegmfile with
        | true -> make_segment_size_file ~link:link ~persist:false ~codec:pref_codec ~max_rep_height:max_rep_height ~extra_cols:extra_cols ~gensegmfile:gensegmfile
        | false -> run_client
            ~link:link
            ~adapt_alg:adapt_alg
            ~initb:initb
            ~maxbuf:maxbuf
            ~persist:persist
            ~turnlogon:turnlogon
            ~logname:logname
            ~v_number:v_number
            ~r_number:r_number
            ~log_folder:subfolder
            ~last_segment_index:last_segment_index
            ~segm_size_from_file:segm_size_from_file
            ~extra_cols:extra_cols
            ~codec:pref_codec
            ~no_segs:no_segs
            ~max_rep_height:max_rep_height
        end
      ]
    )
